
% SaNSDE with restart
% Author:           Mengyuan Zhang
% Created Date:     June 20, 2022
% Modified based on SaNSDE algorithm proposed in:
% Zhenyu Yang, Ke Tang and Xin Yao, "Self-adaptive Differential Evolution with
% Neightborhood Search", in Proceedings of the 2008 IEEE Congress on 
% Evolutionary Computation (CEC2008), Hongkong, China, 2008, pp. 1110-1116.

function [popnew, bestmemnew, bestvalnew, tracerst, ccm, used_FEs, fitMaxVector, gen, yhistory, restart] = sansde_restart(fname, dim_index, pop, val, bestmem, bestval, Lbound, Ubound, itermax, ccm, link_p, caseStudyData,otherParameters, fitMaxVector, gen, yhistory, dorestart, verbose)
    rand('state',sum(otherParameters.iRuns))
    [popsize, dim] = size(pop);
    NP = popsize;
    D = dim;
    tracerst = [];

    %% initialization
    F = zeros(NP,1);

    % initialization for mutation strategy selector
    linkp = link_p;    
    l1 = 1;l2 = 1;nl1 = 1;nl2 = 1;

    % initialization for mutation factor
    fp = 0.5;
    ns1 = 1; nf1 = 1; ns2 = 1; nf2 = 1;

    pm1 = zeros(NP,D);              % initialize population matrix 1
    pm2 = zeros(NP,D);              % initialize population matrix 2
    pm3 = zeros(NP,D);              % initialize population matrix 3
    pm4 = zeros(NP,D);              % initialize population matrix 4
    pm5 = zeros(NP,D);              % initialize population matrix 5
    bm  = zeros(NP,D);              % initialize DE_gbestber  matrix
    ui  = zeros(NP,D);              % intermediate population of perturbed vectors
    mui = zeros(NP,D);              % mask for intermediate population
    mpo = zeros(NP,D);              % mask for old population
    rot = (0:1:NP-1);               % rotating index array (size NP)
    rotd= (0:1:D-1);                % rotating index array (size D)
    rt  = zeros(NP);                % another rotating index array
    rtd = zeros(D);                 % rotating index array for exponential crossover
    a1  = zeros(NP);                % index array
    a2  = zeros(NP);                % index array
    a3  = zeros(NP);                % index array
    a4  = zeros(NP);                % index array
    a5  = zeros(NP);                % index array
    ind = zeros(4);

    % initial for crossover rate adaptation
    cc_rec = [];
    f_rec = [];

    % initial for pop
    gpop = ones(popsize, 1) * bestmem;
    gpop(:, dim_index) = pop;
    [best, ibest] = min(val);
    subbestmem = pop(ibest, :);
    used_FEs = 0;
    iter = 0;
    
    % initial for restart
    restart = 0;
    window = 15;

    %% start main loop
    while iter < itermax
        popold = pop;                   % save the old population
        
        ind = randperm(4);              % index pointer array
        
        a1  = randperm(NP);             % shuffle locations of vectors
        rt = rem(rot+ind(1),NP);        % rotate indices by ind(1) positions
        a2  = a1(rt+1);                 % rotate vector locations
        rt = rem(rot+ind(2),NP);
        a3  = a2(rt+1);                
        rt = rem(rot+ind(3),NP);
        a4  = a3(rt+1);               
        rt = rem(rot+ind(4),NP);
        a5  = a4(rt+1); 
        
        pm1 = popold(a1,:);             % shuffled population 1
        pm2 = popold(a2,:);             % shuffled population 2
        pm3 = popold(a3,:);             % shuffled population 3
        pm4 = popold(a4,:);             % shuffled population 4
        pm5 = popold(a5,:);             % shuffled population 5
        
        bm = ones(NP, 1) * subbestmem;
        
        %% generate crossover rate
        if rem(iter,24)==0
            if (iter~=0) && (~isempty(cc_rec))
                ccm = sum(f_rec.*cc_rec)/sum(f_rec);
            end
            cc_rec = [];
            f_rec = [];
        end
        
        if rem(iter,5)==0
            cc = normrnd(ccm, 0.1, NP*3, 1);
            index = find((cc < 1) & (cc > 0));
            cc = cc(index(1:NP));
        end

        %% generate mutation factor

        fst1 = (rand(NP,1) <= fp);
        fst2 = 1-fst1;

        fst1_index = find(fst1 ~= 0);
        fst2_index = find(fst1 == 0);

        tmp = normrnd(0.5, 0.3, NP, 1);
        F(fst1_index) = tmp(fst1_index);

        tmp = normrnd(0, 1, NP, 1) ./ normrnd(0, 1, NP, 1);
        F(fst2_index) = tmp(fst2_index);

        F = abs(F);
        

        %% prepare corssover. all random numbers < CR are 1, 0 otherwise
        aa = rand(NP,D) < repmat(cc,1,D);
        index = find(sum(aa') == 0);
        tmpsize = size(index, 2);
        for k=1:tmpsize
            bb = ceil(D*rand);
            aa(index(k), bb) = 1;
        end
            
        mui=aa;
        mpo = mui < 0.5;                % inverse mask to mui

        %% mutation strategy selector
        aaa = (rand(NP,1) <= linkp);
        aindex=find(aaa == 0);
        bindex=find(aaa ~= 0);
        
        %% do mutation & crossover
        % use DE/current to best/4
        if ~isempty(bindex)
            % mutation
            ui(bindex,:) = popold(bindex,:)+repmat(F(bindex,:),1,D).*(bm(bindex,:)-popold(bindex,:)) + repmat(F(bindex,:),1,D).*(pm1(bindex,:) - pm2(bindex,:) + pm3(bindex,:) - pm4(bindex,:));
            % crossover
            ui(bindex,:) = popold(bindex,:).*mpo(bindex,:) + ui(bindex,:).*mui(bindex,:);
        end
        % use DE/rand/1
        if ~isempty(aindex)
            ui(aindex,:) = pm3(aindex,:) + repmat(F(aindex,:),1,D).*(pm1(aindex,:) - pm2(aindex,:));
            ui(aindex,:) = popold(aindex,:).*mpo(aindex,:) + ui(aindex,:).*mui(aindex,:);
        end
        bbb=1-aaa; 

        %% Select which vectors are allowed to enter the new population 
        index = find(ui > Ubound);
        ui(index) = Ubound(index);        
        index = find(ui < Lbound);
        ui(index) = Lbound(index);          
        gpop(:, dim_index) = ui;

        %% evaluate new population
        tempval = feval(fname, gpop, caseStudyData,otherParameters);
        yhistory = [yhistory; tempval];

        used_FEs = used_FEs + NP;
        
        %% parameter adaptation
        for i=1:NP
            % for successful offspring
            if (tempval(i) <= val(i))
                % record parameters
                if (tempval(i) < val(i))
                    cc_rec = [cc_rec cc(i,1)];
                    f_rec = [f_rec (val(i) - tempval(i))];
                end
                % update pop and val
                pop(i,:) = ui(i,:);  
                val(i)   = tempval(i);  
                
                l1 = l1 + aaa(i);
                l2 = l2 + bbb(i);

                ns1 = ns1 + fst1(i);
                ns2 = ns2 + fst2(i);
            else
                nl1 = nl1 + aaa(i);
                nl2 = nl2 + bbb(i); 

                nf1 = nf1 + fst1(i);
                nf2 = nf2 + fst2(i);
            end
        end 
        
        % update parameters related to the mutation factor and the stratey
        if (rem(iter,24) == 0) && (iter~=0)
            linkp = (l1/(l1+nl1))/(l1/(l1+nl1)+l2/(l2+nl2));
            l1 = 1;l2 = 1; nl1 = 1; nl2 = 1;
            fp = (ns1 * (ns2 + nf2))/(ns2 * (ns1 + nf1) + ns1 * (ns2 + nf2));
            ns1 = 1; nf1 = 1; ns2 = 1; nf2 = 1;
        end    
        
        %% update and record results

        [best, ibest] = min(val);
        subbestmem = pop(ibest, :);  
        % record best pop
        gen=gen+1;
        fitMaxVector(1,gen) = best;
        
        % update current best
        if (best < bestval)
           bestval = best;
           bestmem(dim_index) = pop(ibest, :);
        end

        tracerst = [tracerst; bestval];
        iter = iter + 1;
        if verbose==1
            fprintf('Fitness value: %f\n',bestval)
            fprintf('Generation: %d\n',iter)
        end

        %% decision making for restart
        if dorestart == 1
            if iter > 55 && iter < 70
                vstd = std(tracerst(end-20:end));
                td = 11;
                if vstd < td
                    fprintf('Stop and restart in phase1_iter%d with std: %.2f thd: %.2f\n', iter, vstd, td);                    
                    restart = 2;
                    break;
                end
            elseif iter > 105 && iter < 120
                vstd = std(tracerst(end-15:end));    
                td = 0.8;  
                if vstd < td
                    fprintf('Stop and restart in phase1_iter%d with std: %.2f thd: %.2f\n', iter, vstd, td);                    
                    restart = 3;
                    break;
                end    
            end
        elseif dorestart == 2
            if iter >55 && iter < 70
                vstd = std(tracerst(end-window:end));
                td = 1.5;
                if vstd < td
                    fprintf('Stop and restart in phase2_iter%d with std: %.2f thd: %.2f\n', iter, vstd, td);
                    restart = 4;
                    break;
                end
            end
        end

    end % end of main loop

    popnew = pop;
    bestmemnew = bestmem;
    bestvalnew = bestval;

end
