clc;clear all;close all;
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% GECAD GECCO and WCCI 2022 Competition: Evolutionary Computation in the Energy Domain: Risk-based Energy Scheduling
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
addpath('functions') %Necessary functions to run the algorithms (encrypted)
tTotalTime=tic; % lets track total computational time
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Select testbed
Select_testbed=1; %Only 1 track in 2022
%Testbed 1: Risk-based energy scheduling    
DB=Select_testbed;
% 1: Case study testbed 1

Select_algorithm=2;
%1: DE algorithm (test algorithm)
%2: Your algorithm

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Load MH parameters (e.g., get MH parameters from DEparameters.m file)
switch Select_algorithm
     case 1
          addpath('DEalg')
          algorithm='DE-rand'; %'The participants should include their algorithm here'
          DEparameters %Function defined by the participant
          No_solutions=deParameters.I_NP; %Notice that some algorithms are limited to one individual
        
        case 2
          %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
          %Your algorithm can be inserted here
          %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
          addpath('ReSaDEalg')
          algorithm='ReSaDe'; %'The participants should include their algorithm here'
          DEparameters %Function defined by the participant
%           No_solutions=deParameters.I_NP;
          No_solutions=10;
          deParameters.F_weight= 0.8986;
          deParameters.F_CR= 0.1375;
    otherwise
          fprintf(1,'No algorithm selected\n');
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Load Data base 
[caseStudyData, DB_name]=callDatabase(DB);
noRuns=20; %this can be changed but final results should be based on 20 trials

%% Label of the algorithm and the case study
Tag.algorithm=algorithm;
Tag.DB=DB_name;
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Set other parameters
otherParameters =setOtherParameters(caseStudyData, No_solutions, Select_testbed);
%% Set lower/upper bounds of variables 
[lowerB,upperB] = setVariablesBounds(caseStudyData,otherParameters);
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Call the MH for optimizationclear 
ResDB=struc([]);
yhistory_all = [];

for iRuns=1:noRuns %Number of trails
       tOpt=tic;
       rand('state',sum(noRuns*100*clock))% ensure stochastic indpt trials
       otherParameters.iRuns=iRuns;      
        
       switch Select_algorithm
              case 1
                  [ResDB(iRuns).Fit_and_p, ...
                  ResDB(iRuns).sol, ...
                  ResDB(iRuns).fitVector]= ...                    
                  deopt_simple(deParameters,caseStudyData,otherParameters,lowerB,upperB);  
              
              case 2     
                    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                    % Your algorithm
                    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                  [ResDB(iRuns).Fit_and_p, ...
                  ResDB(iRuns).sol, ...
                  ResDB(iRuns).fitVector, yhistory]= ...                    
                  ReSaDE(deParameters,caseStudyData,otherParameters,lowerB,upperB,iRuns);
                  yhistory_all = [yhistory_all; yhistory];
       end 
            
            ResDB(iRuns).tOpt=toc(tOpt); % time of each trial
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            %% Save the results and stats
            Save_results
        
end
save('./results/y_res.mat', 'yhistory_all');
save('./results/ResDB.mat', 'ResDB');

%% End of MH Optimization
% for j=1:noRuns
%       lower_violations=ResDB(j).sol<lowerB;
%       upper_violations=ResDB(j).sol>upperB;
%       many_low(j)=sum(lower_violations);
%       many_up(j)=sum(upper_violations);
%       fprintf('lower_violations : %d .\n', many_low(j));
%       fprintf('upper_violations : %d .\n', many_up(j));
% end

tTotalTime=toc(tTotalTime); %Total time